/*
	-------------------------------------------
	 C Memory management surveillance routines
	-------------------------------------------

	Author: Martin Vorlaender (mv@pdv-systeme.de)
	        based on code written by
	        Diethard Wettrau  (dw@pdv-systeme.de)

 This tool redirects all calls to standard C and X memory management
 routines (see MEMTRC.H for a list) to its own routines, which keep a
 list of all allocated and freed memory blocks. On a call to flush_memlist(),
 the list is written to a file or stderr. Calling a free routine
 with either an address not previously used in an allocation routine call
 or an address already used in a free routine call immediately triggers an
 output to stderr.

 All of this is disabled if NDEBUG is defined.

 To get this tool working, #include MEMTRC.H into all your modules
 (_after_ all system includes) and insert a call to memtrc_flush_list()
 before exiting your program. Use memtrc_notify_alloc() to tell the
 memory list management about non-redirectable memory allocations.


 Prototypes for the interface routines are

 extern void memtrc_flush_list ( char *log_file_name ); # if NULL: use stderr
 extern void memtrc_notify_alloc ( void *addr, size_t size );


 Compile-time options (using C macro definitions) :

  USE_C       Use standard C memory management routines (Default)
   or
  USE_XT      Use X Toolkit memory management routines
   or
  USE_OPT     Use VMS memory management routines (VMS only)

  KEEP_ALL    If defined to 1, memory list items are not dismissed on
              deallocation, i.e. memtrc_flush_list() gets you the complete
              list of all memory allocations
              (Default: 0)

  MAX_TRIES   Maximum number of tries on any malloc()/calloc()
              (Default: 3)

  SHOW_VM     If defined to 1, on every call to a routine of this module
              the virtual memory status is checked, and statistics are
              printed.
              (VMS only; Default: 0)

  CHECK_VM    Same as SHOW_VM, but statistics are printed only if
              the check fails
              (VMS only; Default: 1)

  FENCE_SIZE  Minimum size (in bytes) of the fenceposts guarding
              the integrity of the allocated memory block
              (Default: 8)

  FENCE_VALUE Pattern to fill the fencepost bytes with
              (Default: 0x55)
*/

/*--------------- no user serviceable parts beyond this point ---------------*/

#ifndef NDEBUG

#if defined(VMS) || defined(__VMS)
#	define IS_VMS 1
#else
#	define IS_VMS 0
#endif /* VMS */

/* USE_OPT only is available on VMS */
#if defined(USE_OPT) && !IS_VMS
#	undef USE_OPT
#	define USE_C
#endif /* USE_OPT && !VMS */

/* Default for memory allocation routine */
#if !defined(USE_XT) && !defined(USE_OPT) && !defined(USE_C)
#	define USE_C
#endif /* !USE_XT && !USE_OPT && !USE_C */

/* Default for KEEP_ALL */
#ifndef KEEP_ALL
#	define KEEP_ALL 0
#endif /* !KEEP_ALL */

/* Default for MAX_TRIES */
#ifndef MAX_TRIES
#	define MAX_TRIES 3
#endif /* !MAX_TRIES */

/* Default for SHOW_VM and CHECK_VM (VMS only) */
#if IS_VMS
#	ifndef SHOW_VM
#		define SHOW_VM 0
#	endif /* !SHOW_VM */
#	if SHOW_VM
#		define CHECK_VM 1
#	else
#		ifndef CHECK_VM
#			define CHECK_VM 1
#		endif /* !CHECK_VM */
#	endif /* SHOW_VM != 0 */
#else
#	define SHOW_VM  0
#	define CHECK_VM 0
#endif /* IS_VMS */

/* Default for FENCE_SIZE */
#ifndef FENCE_SIZE
#	define FENCE_SIZE 8
#endif /* FENCE_SIZE */

/* Default for FENCE_VALUE */
#ifndef FENCE_VALUE
#	define FENCE_VALUE 0x55
#endif /* FENCE_VALUE */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#ifndef FALSE
#	define FALSE 0
#	define TRUE  1
#endif

#if IS_VMS
#	pragma nostandard

#	if __VMS_VER >= 70000000
#		define __NEW_STARLET
#		define USE_STD_LIB_HEADER
#	endif /* __VMS_VER >= 70000000 */

#	include ssdef
#	include descrip

#	ifdef USE_STD_LIB_HEADER
#		include lib$routines
#	else
		extern unsigned long lib$verify_vm_zone
		(
			unsigned long *zone_id
		);
		extern unsigned long lib$show_vm
		(
			long          *code,
			unsigned long (*user_action_procedure)(),
			unsigned long user_specified_argument
		);
		extern unsigned long lib$show_vm_zone
		(
			unsigned long *zone_id,
			long          *detail_level,
			unsigned long (*user_action_procedure)(),
			unsigned long user_arg
		);
#	endif /* USE_STD_LIB_HEADER */

#	if defined(USE_OPT) && defined(__DECC)
		extern void* VAXC$MALLOC_OPT( size_t size );
		extern void* VAXC$CALLOC_OPT( size_t items, size_t size );
		extern void* VAXC$REALLOC_OPT( void *addr, size_t size );
		extern void  VAXC$FREE_OPT( void *addr );
		extern void  VAXC$CFREE_OPT( void *addr );
#	endif /* USE_OPT && __DECC */
#endif /* IS_VMS */

#ifdef USE_XT
#	include <X11/Intrinsic.h>
#endif /* USE_XT */

#if IS_VMS
#	pragma standard
#endif /* IS_VMS */

/***********************************************************************/

#define NO_LINE -1

typedef struct memtrc_item
{
	void   *addr;
	size_t size;
	size_t fence_size;

	char   *alloc_file;
	long   alloc_line;   /* or NO_LINE */
	long   alloc_serial;

	char   *free_file;
	long   free_line;    /* or NO_LINE */
	long   free_serial;

	struct memtrc_item *prev;
	struct memtrc_item *next;
} item, *itemPtr;


static itemPtr itemListRoot = NULL,
               itemListTail = NULL;

static long itemList_serial = 0;

static long alloc_siz     = 0,
            free_siz      = 0,
            max_alloc_siz = 0;


static void *internal_malloc
(
	size_t size
)
{
#	ifdef USE_XT
		return XtMalloc( size );
#	else
#		ifdef USE_OPT
			return VAXC$MALLOC_OPT( size );
#		else /* must be USE_C */
			return malloc( size );
#		endif /* USE_OPT */
#	endif /* USE_XT */
}


static void *internal_calloc
(
	size_t items,
	size_t size
)
{
#	ifdef USE_XT
		return XtCalloc( items, size );
#	else
#		ifdef USE_OPT
			return VAXC$CALLOC_OPT( items, size );
#		else /* must be USE_C */
			return calloc( items, size );
#		endif /* USE_OPT */
#	endif /* USE_XT */
}


static void *internal_realloc
(
	void   *addr,
	size_t size
)
{
#	ifdef USE_XT
		return XtRealloc( addr, size );
#	else
#		ifdef USE_OPT
			return VAXC$REALLOC_OPT( addr, size );
#		else /* must be USE_C */
			return realloc( addr, size );
#		endif /* USE_OPT */
#	endif /* USE_XT */
}


static void internal_free
(
	void *addr
)
{
#	ifdef USE_XT
		XtFree( addr );
#	else
#		ifdef USE_OPT
			VAXC$FREE_OPT( addr );
#		else /* must be USE_C */
			free( addr );
#		endif /* USE_OPT */
#	endif /* USE_XT */
}


static char *internal_strdup
(
	char *src
)
{
	char *dst;

	if (src == NULL)
		return NULL;

	dst = internal_malloc( strlen(src) +1 );
	assert( dst != NULL );

	strcpy( dst, src );

	return dst;
}


static void free_and_null
(
	char **src
)
{
	if (*src == NULL)
		return;

	internal_free( *src );

	*src = NULL;
}


#if IS_VMS
	static unsigned long check_vm_action
	(
		struct dsc$descriptor_s *resultant_string,
		unsigned long           user_specified_argument
	)
	{
		fprintf
			(
			stderr, "%.*s\n",
			(unsigned short)resultant_string->dsc$w_length,
			resultant_string->dsc$a_pointer
			);
		return SS$_NORMAL;
	}


	static void check_vm
	(
		char   *called_by,

		char   *source,
		long   line,

		size_t size
	)
	{
		unsigned long Zone = 0; /* Default zone */
		unsigned long RC;
		long          code;

		RC = lib$verify_vm_zone( &Zone );
		if (!(RC & 1) || SHOW_VM != 0)
		{
			if (!(RC & 1))
				fprintf( stderr, "VM zone verification failed on %s with condition %lu\n",
				         called_by, RC );
			else /* SHOW_VM != 0 */
				fprintf( stderr, "VM zone status (called by %s)\n", called_by );

			fprintf( stderr, "  source : %s\n",  source );
			fprintf( stderr, "  line   : %ld\n", line );
			fprintf( stderr, "  size   : %ld\n", size );
			fprintf( stderr, "  serial : %ld\n", itemList_serial+1 );

			code = 0; lib$show_vm( &code, check_vm_action, 0 );
			code = 4; lib$show_vm( &code, check_vm_action, 0 );
			code = 2; lib$show_vm_zone( &Zone, &code, check_vm_action, 0 );
		}
	}
#endif /* IS_VMS */


static void append_item
(
	void   *addr,
	size_t size,
	size_t fence_size,

	char   *alloc_file,
	long   alloc_line,

	char   *free_file,
	long   free_line
)
{
	itemPtr pItem;

	assert( alloc_line != NO_LINE || free_line != NO_LINE);

	pItem = (itemPtr) internal_malloc( sizeof(item) );
	assert( pItem != NULL );

	pItem->addr       = addr;
	pItem->size       = size;
	pItem->fence_size = fence_size;

	pItem->alloc_file = internal_strdup( alloc_file );
	pItem->alloc_line = alloc_line;

	pItem->free_file  = internal_strdup( free_file );
	pItem->free_line  = free_line;

	if (alloc_line != NO_LINE)
		pItem->alloc_serial = ++itemList_serial;
	else /* (free_line != NO_LINE) */
		pItem->free_serial  = ++itemList_serial;

	if (itemListRoot == NULL)
	{
		pItem->prev = NULL;

		itemListRoot =
		itemListTail = pItem;
	}
	else
	{
		pItem->prev = itemListTail;

		itemListTail       =
		itemListTail->next = pItem;
	}
	pItem->next = NULL;
}


static void report_fence_status
(
	itemPtr pItem,
	FILE    *file,
	int     always_report
)
{
	int  stat;

	if (pItem->fence_size == 0)
	{
		stat = -1;
	}
	else
	{
		char *lwr = (char*)pItem->addr,
		     *upr = (char*)pItem->addr + pItem->fence_size + pItem->size;
		int  i;

		stat = 0;
		for (i = 0; i < pItem->fence_size; ++i)
		{
			if (*lwr++ != FENCE_VALUE) stat |= 1;
			if (*upr++ != FENCE_VALUE) stat |= 2;
		}
	}

	switch (stat)
	{
		case -1 :
			if (always_report)
				fprintf( file, "Fencepost status : (not fenceposted)\n" );
			break;
		case 0 :
			if (always_report)
				fprintf( file, "Fencepost status : ok\n" );
			break;
		case 1 :
			fprintf( file, "Fencepost status : lower compromized, upper ok!\n" );
			break;
		case 2 :
			fprintf( file, "Fencepost status : lower ok, upper compromized!\n" );
			break;
		case 3 :
			fprintf( file, "Fencepost status : lower and upper compromized!\n" );
			break;
	}
}


static void flush_itemList
(
	char *log_file_name
)
{
	FILE    *file;
	itemPtr pItem;

	file = (log_file_name != NULL) ? fopen( log_file_name, "w" ) : stderr;
	assert( file != NULL );

	fprintf( file, "\nMemory allocation and deallocation protocol" );
	fprintf( file, "\n-------------------------------------------\n" );

	fprintf( file, "Max. allocated  : %15ld bytes\n", max_alloc_siz );
	fprintf( file, "Max. serial     : %15ld\n\n", itemList_serial );

	fprintf( file, "Allocated       : %15ld bytes\n", alloc_siz );
	fprintf( file, "Deallocated     : %15ld bytes\n", free_siz );
	fprintf( file, "---------------------------------------\n" );
	fprintf( file, "Balance         : %+15ld bytes\n\n", alloc_siz - free_siz );

	for (pItem = itemListRoot; pItem != NULL; pItem = pItem->next)
	{
		fprintf( file, "-------------------------------------------------------\n" );

		fprintf( file, "Address          : 0x%lx\n", pItem->addr );

		fprintf( file, "Size (bytes)     : " );
		if (pItem->size != 0)
			fprintf( file, "%d\n", pItem->size );
		else
			fprintf( file, "?\n" );

		if (pItem->alloc_line == NO_LINE)
			fprintf( file, "Not Allocated!\n" );
		else
		{
			fprintf( file, "Allocated\n" );
			fprintf( file, "  in source file : %s\n",  pItem->alloc_file   );
			fprintf( file, "         in line : %ld\n", pItem->alloc_line   );
			fprintf( file, "          serial : %ld\n", pItem->alloc_serial );
		}

		if (pItem->free_line == NO_LINE)
			fprintf( file, "Not Deallocated!\n" );
		else
		{
			fprintf( file, "Deallocated\n" );
			fprintf( file, "  in source file : %s\n",  pItem->free_file   );
			fprintf( file, "         in line : %ld\n", pItem->free_line   );
			fprintf( file, "          serial : %ld\n", pItem->free_serial );
		}

		report_fence_status( pItem, file, TRUE );
	}

	if (file != stderr)
		fclose( file );
}


static void dismiss_itemList
(
	void
)
{
	itemPtr pItem,
	        pNext;

	for (pItem = itemListRoot; pItem != NULL; pItem = pNext)
	{
		pNext = pItem->next;

		free_and_null( &pItem->alloc_file );
		free_and_null( &pItem->free_file  );
		internal_free( pItem );
	}

	itemListRoot =
	itemListTail = NULL;

	itemList_serial = 0;

	alloc_siz     =
	free_siz      =
	max_alloc_siz = 0;
}


static void add_alloc_item
(
	void   *addr,
	size_t size,
	size_t fence_size,

	char   *source,
	long   line
)
{
	append_item( addr, size, fence_size, source, line, NULL, NO_LINE );

	alloc_siz += size;

	if (alloc_siz - free_siz > max_alloc_siz)
		max_alloc_siz = alloc_siz - free_siz;
}


static itemPtr find_item
(
	void *addr
)
{
	itemPtr pItem;

	for (pItem = itemListRoot; pItem != NULL; pItem = pItem->next)
		if ((char*)pItem->addr + pItem->fence_size == addr)
			break;
	return pItem;
}


static void add_free_item
(
	void          *addr,

	char          *source,
	unsigned long line
)
{
	itemPtr pItem;

	pItem = find_item( addr );
	if (pItem != NULL)
	{
		if (pItem->free_line != NO_LINE)
		{
			fprintf( stderr, "--------------------------------\n"          );
			fprintf( stderr, "Multiple freeing of memory block\n"          );
			fprintf( stderr, "source          : %s\n" , source             );
			fprintf( stderr, "line            : %ld\n", line               );
			fprintf( stderr, "previous source : %s\n" , pItem->free_file   );
			fprintf( stderr, "previous line   : %ld\n", pItem->free_line   );
			fprintf( stderr, "previous serial : %ld\n", pItem->free_serial );
			fprintf( stderr, "--------------------------------\n"          );
		}
		else
		{
			free_siz += pItem->size;
	
#			if KEEP_ALL
				pItem->free_file   = internal_strdup( source );
				pItem->free_line   = line;
				pItem->free_serial = ++itemList_serial;
#			else
				report_fence_status( pItem, stderr, FALSE );

				free_and_null( &pItem->alloc_file );
				free_and_null( &pItem->free_file  );
	
				if (pItem == itemListRoot && pItem == itemListTail)
				{
					itemListRoot =
					itemListTail = NULL;
				}
				else if (pItem == itemListRoot)
				{
					itemListRoot = pItem->next;
					itemListRoot->prev = NULL;
				}
				else if (pItem == itemListTail)
				{
					itemListTail = pItem->prev;
					itemListTail->next = NULL;
				}
				else
				{
					pItem->prev->next = pItem->next;
					pItem->next->prev = pItem->prev;
				}

				internal_free( pItem );
#			endif
		}
	}
	else /* !found */
	{
		fprintf( stderr, "----------------------------\n"    );
		fprintf( stderr, "Freeing unknown memory block\n"    );
		fprintf( stderr, "source : %s \n", source            );
		fprintf( stderr, "line   : %ld\n", line              );
		fprintf( stderr, "serial : %ld\n", itemList_serial+1 );
		fprintf( stderr, "----------------------------\n"    );

		append_item( addr, 0, 0, NULL, NO_LINE, source, line );
	}
}


static void memtrc_free2
(
	void **addrPtr,

	char *file,
	long line,

	char *called_by
)
{
	if (*addrPtr == NULL)
		return;

	internal_free( *addrPtr );

#	if CHECK_VM
		check_vm( called_by, file, line, 0 );
#	endif /* CHECK_VM */

	add_free_item( *addrPtr, file, line );

	*addrPtr = NULL;
}

/***********************************************************************/


void *memtrc_malloc
(
	size_t size,

	char   *file,
	long   line
)
{
	char *ptr;
	int  i;

	for (i = 0; i < MAX_TRIES; ++i)
	{
		ptr = (char*) internal_malloc( size + 2*FENCE_SIZE );
		if (ptr != NULL)
		{
			memset( ptr                , FENCE_VALUE, FENCE_SIZE );
			memset( ptr+FENCE_SIZE+size, FENCE_VALUE, FENCE_SIZE );
			break;
		}

		fprintf( stderr, "malloc failed on try %d\n", i+1 );
		if (i == 0)
		{
			fprintf( stderr, "source : %s \n", file );
			fprintf( stderr, "line   : %ld\n", line );
			fprintf( stderr, "size   : %ld\n", size );
		}
	}

#	if CHECK_VM
		check_vm( "malloc", file, line, size );
#	endif /* CHECK_VM */

	add_alloc_item( ptr, size, FENCE_SIZE, file, line );

	return (void*)(ptr + FENCE_SIZE);
}


void *memtrc_calloc
(
	size_t items,
	size_t size,

	char   *file,
	long   line
)
{
	size_t fence_items;
	size_t fence_size;
	char   *ptr;
	int    i;

	assert( size != 0 );
	fence_items = FENCE_SIZE / size + ((FENCE_SIZE % size != 0) ? 1 : 0);
	fence_size  = fence_items * size;

	for (i = 0; i < MAX_TRIES; ++i)
	{
		ptr = (char*) internal_calloc( items+2*fence_items, size );
		if (ptr != NULL)
		{
			memset( ptr                         , FENCE_VALUE, fence_size );
			memset( ptr+(fence_items+items)*size, FENCE_VALUE, fence_size );
			break;
		}

		fprintf( stderr, "calloc failed on try %d\n",i );
		if (i == 0)
		{
			fprintf( stderr, "source : %s \n", file );
			fprintf( stderr, "line   : %ld\n", line );
			fprintf( stderr, "items  : %ld\n", items );
			fprintf( stderr, "size   : %ld\n", size );
		}
	}

#	if CHECK_VM
		check_vm( "calloc", file, line, size * items );
#	endif /* CHECK_VM */

	add_alloc_item( ptr, size * items, fence_size, file, line );

	return (void*)(ptr + fence_size);
}


void *memtrc_realloc
(
	void   **addrPtr,
	size_t size,

	char *file,
	long line
)
{
	itemPtr pItem;
	size_t  old_size = 0;
	void    *old_addr = NULL;
	char    *ptr;
	int     i;

	pItem = find_item( *addrPtr );
	if (pItem != NULL)
	{
		old_addr = pItem->addr;
		old_size = pItem->size;
	}
	add_free_item( *addrPtr, file, line );

	for (i = 0; i < MAX_TRIES; ++i)
	{
		ptr = (char*) internal_realloc( old_addr, size + 2*FENCE_SIZE );
		if (ptr != NULL)
		{
			memset( ptr                , FENCE_VALUE, FENCE_SIZE );
			memset( ptr+FENCE_SIZE+size, FENCE_VALUE, FENCE_SIZE );
			break;
		}

		fprintf( stderr, "realloc failed on try %d\n", i+1 );
		if (i == 0)
		{
			fprintf( stderr, "source : %s \n", file );
			fprintf( stderr, "line   : %ld\n", line );
			fprintf( stderr, "size   : %ld\n", size );
		}
	}

#	if CHECK_VM
		check_vm( "realloc", file, line, size );
#	endif /* CHECK_VM */

	add_alloc_item( ptr, size, FENCE_SIZE, file, line );

	return (void*)(ptr + FENCE_SIZE);
}


char *memtrc_strdup
(
	char *str,

	char *file,
	long line
)
{
	char *ptr;

	if (str == NULL)
		return NULL;

	ptr = (char*) memtrc_malloc( strlen(str)+1, file, line );
	if (ptr != NULL)
		strcpy( ptr, str );

	return ptr;
}


void memtrc_free
(
	void **addrPtr,

	char *file,
	long line
)
{
	memtrc_free2( addrPtr, file, line, "free" );
}


void memtrc_cfree
(
	void **addrPtr,

	char *file,
	long line
)
{
	memtrc_free2( addrPtr, file, line, "cfree" );
}


void memtrc_notify_alloc2
(
	void   *addr,
	size_t size,

	char   *file,
	long   line
)
{
	add_alloc_item( addr, size, 0, file, line );
}

void memtrc_flush_list
(
	char *log_file_name
)
{
	flush_itemList( log_file_name );
	dismiss_itemList();
}

#endif /* NDEBUG */
